/*
** Copyright (c) 1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/


#include "URomFlasher.h"

#include <DriverServices.h>
#include <PCI.h>

#include <3dfx.h>
#include <gdebug.h>
#include <h3cinit.h>
#include <h3regs.h>
#include <h3cinitdd.h>
#include <minihwc.h>
#include <hwcio.h>


#define kPCIPhysicalAddressProperty "AAPL,address"



void				GetROMID(
							TRomInfo *		inRomInfo );

void				ReadVoodooMapping(
 							TRomInfo *		inRomInfo);
 							
PLATIDT				GetROMType(
							TRomBuffer *    inBuffer);

void				ROMErase(
							TRomInfo *		inRomInfo);
							
FLASHROMERROR		ROMWriteBuffer(
							TRomInfo *		inRomInfo);
							
void				SendROMCommand(
							TRomInfo *		inRomInfo,
							FxU8            inCommand);
							
void				BeginROM(
							TRomInfo *		inRomInfo);

void				EndROM(
							TRomInfo *		inRomInfo);

FxU8				ROMByteRead(
							TRomInfo *		inRomInfo,
							FxU16           inOffset);

void				ROMByteWrite (
							TRomInfo *		inRomInfo,
							FxU16           inOffset,
							FxU8            inData);

FxU8				ROMVerifyWrite (
							TRomInfo *		inRomInfo,
							FxU16			inOffset,
							FxU8			inExpected);
							
void				FlushBus(
							TRomInfo *		inRomInfo);

void				Delay15us(
							int		        inCount);




/*______________________________________________________________________________

FlashROM


on entry, the following params must be valid:
  mRegEntryID
  mBuffer->mBuffer
  mBuffer->mBufferSize
  
*/

FLASHROMERROR
FlashROM(
  TRomInfo *		inRomInfo)
{
  FLASHROMERROR     theSuccess = kFRE_unkownError;
  FxU32				theImageType;
  
  ReadVoodooMapping( inRomInfo );
  EnableVoodoo( inRomInfo );

  BeginROM( inRomInfo );

  GetROMID( inRomInfo );	
  if ( inRomInfo->mSectorSize < 0 )
  {
    theSuccess = kFRE_unkownROMDevice;
    goto ErrorExit;
  }

  /* Figure out if it's a PC or Mac ROM image. */	
  theImageType = GetROMType( inRomInfo->mBuffer );
    
  if(theImageType != 1) {
    if ( theImageType == 0 )  /* this is a PC BIOS ROM */
      theSuccess = kFRE_cannotFlashApcROM;
    else                  /* the content of this ROM is unkown */
      theSuccess = kFRE_cannotFlashUnkownROM;
    
    goto ErrorExit;
  }

    
  /* Write the ROM image */
  ROMErase( inRomInfo );
  theSuccess = ROMWriteBuffer( inRomInfo );


ErrorExit:
	// Restore the PCI state to its original state
	EndROM( inRomInfo );

	return theSuccess;
}





/*______________________________________________________________________________

ReadVoodooMapping


*/

void
ReadVoodooMapping(
  TRomInfo *		inRomInfo)
{
#define kNumAssignedAddresses 4
  RegEntryID
    theRegEntryID;
  hwcRegInfo *
    theRegInfo = &inRomInfo->mRegInfo;

  PCIAssignedAddress
    theAssignedAddresses[kNumAssignedAddresses];
  UInt32
    theApplAddress[kNumAssignedAddresses];
  RegPropertyValueSize
    thePropSize;
  UInt32
    theReg0Index,
    theReg1Index,
    theReg2Index,
    theReg3Index;
  OSErr
    theSuccess = noErr;
  short
    i;

  RegistryEntryIDCopy( inRomInfo->mRegEntryID, &theRegEntryID );

  theRegInfo->initialized = FXFALSE;
  
  theSuccess = RegistryPropertyGetSize( &theRegEntryID, kPCIAssignedAddressProperty, &thePropSize);
  if((theSuccess == noErr) && (thePropSize <= sizeof(theAssignedAddresses))) {
          
    /* Grab the property */
    theSuccess = RegistryPropertyGet( &theRegEntryID, kPCIAssignedAddressProperty, theAssignedAddresses, &thePropSize);
    if(theSuccess == noErr) {
          
      /*  This is done because the order of the entries is the assigned addresses property do not necessarily
          come in the same order as the hardware registers. */  
      for(i = 0; i < kNumAssignedAddresses; i++) {
        switch ( theAssignedAddresses[i].registerNumber ) {
          case 0x10:
            theReg0Index = i;
            break;
          
          case 0x14:
            theReg1Index = i;
            break;

          case 0x18:
            theReg2Index = i;
            break;

          case 0x30:
            theReg3Index = i;
            break;
        }  
      }
          
      theSuccess = RegistryPropertyGetSize( &theRegEntryID, kPCIPhysicalAddressProperty, &thePropSize);
      if((theSuccess == noErr) && (thePropSize <= sizeof(theApplAddress))) {
            
        theSuccess = RegistryPropertyGet( &theRegEntryID, kPCIPhysicalAddressProperty, theApplAddress, &thePropSize);
        if(theSuccess == noErr) {
          FxU32
            theHwBaseAddress;
          
          theHwBaseAddress = theApplAddress[theReg0Index];
 
          inRomInfo->mHwBaseAddress = theHwBaseAddress;
          
          theRegInfo->initialized = FXTRUE;
          theRegInfo->ioMemBase = theHwBaseAddress + SST_IO_OFFSET;
          theRegInfo->cmdAGPBase = theHwBaseAddress + SST_CMDAGP_OFFSET;
          theRegInfo->waxBase = theHwBaseAddress + SST_2D_OFFSET;
          theRegInfo->sstBase = theHwBaseAddress + SST_3D_OFFSET;
          theRegInfo->lfbBase = theHwBaseAddress + SST_LFB_OFFSET;
          theRegInfo->rawLfbBase = (theApplAddress[theReg1Index] & ~0xf);
          theRegInfo->ioPortBase = (theApplAddress[theReg2Index] & ~0xf);

          inRomInfo->mOriginalRomBaseAddr = theApplAddress[theReg3Index];

        }
      }
    }
  }
}


/*______________________________________________________________________________

EnableVoodoo3

Enable PCI Memory space and IO space access. It should have already been done,
but just in case...

*/

void
EnableVoodoo(
  TRomInfo *		inRomInfo)
{
  FxU32 
    thePciInit0;

    ExpMgrConfigWriteLong( inRomInfo->mRegEntryID, (LogicalAddress)0x4, 3);
  
    /* slow down PCI accesses */
    HWC_IO_LOAD( (inRomInfo->mRegInfo), pciInit0, thePciInit0);
    thePciInit0 |= SST_PCI_READ_WS | SST_PCI_WRITE_WS;
    HWC_IO_STORE( (inRomInfo->mRegInfo), pciInit0, thePciInit0);  
}





/*______________________________________________________________________________

ReadRomInfo

*/

void
ReadRomInfo(
  TRomInfo *		ioRomInfo)
{
    
  BeginROM( ioRomInfo );

  GetROMID( ioRomInfo );
	
  if ( ioRomInfo->mSectorSize < 0 )
  {
    short i;
    FxU8 * theBPtr = ioRomInfo->mBuffer->mBuffer;
    for(i = 0; i < ioRomInfo->mBuffer->mBufferSize; i++)
    {
      theBPtr[i] = ROMByteRead( ioRomInfo, i);
    }
    
 	ioRomInfo->mPlatform = GetROMType( ioRomInfo->mBuffer );

  }

  EndROM( ioRomInfo );
}





/*______________________________________________________________________________

GetROMID - Identifies the ROM manufacturer and device ID

Entry:	lpMan               Pointer to the manufacturer ID (returned)
        lpDevice            Pointer to the device ID (returned)
        lpwSectorSize       Size (in BYTEs) of an EEPROM sector
        
Exit:   None

*/

void GetROMID(
  TRomInfo *		inRomInfo )
{
  SendROMCommand( inRomInfo, 0xF0);									/* Exit ID mode (just in case) */
  SendROMCommand( inRomInfo, 0x90);									/* Enter ID mode */
  inRomInfo->mManufacturer = (MANIDT) ROMByteRead( inRomInfo, 0);	/* Read manufacturer ID */
  inRomInfo->mDevice = (DEVIDT) ROMByteRead( inRomInfo, 1);			/* Read device ID */
  SendROMCommand( inRomInfo, 0xF0);									/* Exit ID mode */

  switch ( inRomInfo->mManufacturer )
  {
    case MANID_AMD_TI:

      inRomInfo->mSectorSize = 1;
      break;



    case MANID_ATMEL:
		
      if( inRomInfo->mDevice == 3 )
        inRomInfo->mDevice = ATMEL_49F010;
				
      if( inRomInfo->mDevice == ATMEL_49F010 )
        inRomInfo->mSectorSize = 1;
      else
        inRomInfo->mSectorSize = 128;
      break;



    case MANID_SST:

      if( inRomInfo->mDevice == SST_29EE010 )
        inRomInfo->mSectorSize = 128;
      else
        inRomInfo->mSectorSize = 1;
      break;



    default:

      inRomInfo->mSectorSize = 1;
      break;
  }


  if ( !(((inRomInfo->mManufacturer == MANID_AMD_TI) && (inRomInfo->mDevice == AMD_TI_29010)) ||
         ((inRomInfo->mManufacturer == MANID_ATMEL) && (inRomInfo->mDevice == ATMEL_29010)) ||
		 ((inRomInfo->mManufacturer == MANID_ATMEL) && (inRomInfo->mDevice == ATMEL_49F010)) ||
		 ((inRomInfo->mManufacturer == MANID_SST) && (inRomInfo->mDevice == SST_29EE010)) ||
		 ((inRomInfo->mManufacturer == MANID_SST) && (inRomInfo->mDevice == SST_39SF010)) ||
		 ((inRomInfo->mManufacturer == MANID_SST) && (inRomInfo->mDevice == SST_39VF512)) )
		)
  {
    inRomInfo->mSectorSize = -1;
  }
  
}





/*______________________________________________________________________________

GetROMType

*/

PLATIDT
GetROMType(
  TRomBuffer *      inBuffer)
{
  FxU16 theHeaderOffset = 0;
  FxU8 * theBuffer = inBuffer->mBuffer;
  PLATIDT thePlatform = PLATID_UNKNOWN;
  
  /* First make sure we have a valid PCI header. */
  if( theBuffer[theHeaderOffset + 0] == 0x55 && theBuffer[theHeaderOffset + 1] == 0xAA) {
    FxU16 theImageOffset;
    
    /* Get offset to first image's PCI Data Structure */
    theImageOffset = theBuffer[theHeaderOffset + 0x18] | (theBuffer[theHeaderOffset + 0x19] << 8);
    
    /* Return code type byte */
    thePlatform = (PLATIDT) theBuffer[theImageOffset + 0x14]; 
    
  }
  return thePlatform;
}





/*______________________________________________________________________________

ROMErase

*/

void ROMErase(
  TRomInfo *		inRomInfo)
{
	SendROMCommand( inRomInfo, 0x80 );
	SendROMCommand( inRomInfo, 0x10 );

	/* Wait for the ROM to erase itself */

	while( (ROMByteRead( inRomInfo,0 ) & 0x80) == 0x0)
		;


	/* Play it safe and wait for 10ms */
	Delay15us( 20000l/15 );
}





/*______________________________________________________________________________

ROMWriteBuffer

*/

FLASHROMERROR ROMWriteBuffer(
  TRomInfo *		inRomInfo)
{
  FLASHROMERROR     theSuccess = kFRE_unkownError;
  FxU16				wCounter;
  FxU8 *            theRomDataPtr;
  unsigned int		i;

  // determine that the ROM has the correct signature and ROM size.
  theRomDataPtr = inRomInfo->mBuffer->mBuffer;
  if (!((theRomDataPtr[0] == 0x55) && (theRomDataPtr[1] == 0xAA)))
  {
    theSuccess = kFRE_badROMHeader;
    goto ErrorExit;
  }

  // If processing reaches here, then a valid ROM image is available for
  // downloading to the EEPROM. Rewind the file, and start reading again.
  wCounter = inRomInfo->mSectorSize;
  for (i = 0; i < inRomInfo->mBuffer->mBufferSize ; i++)
  {
    if (wCounter >= inRomInfo->mSectorSize)
    {
      SendROMCommand( inRomInfo, 0xA0 );
      wCounter = 0;
    }

    ROMByteWrite( inRomInfo, i, theRomDataPtr[i]);
    if (++wCounter >= inRomInfo->mSectorSize)
    {
      if (!ROMVerifyWrite( inRomInfo, i, theRomDataPtr[i] ) )
      {
        theSuccess = kFRE_FailedDuringProgramming;
        goto ErrorExit;
      }
    }
  }

  theSuccess = kFRE_noErr;
  
ErrorExit:
	return theSuccess;
}





/*______________________________________________________________________________

SendROMCommand

*/

void
SendROMCommand(
  TRomInfo *		inRomInfo,
  FxU8              inCommand)
{
  ROMByteWrite( inRomInfo, 0x5555, 0xAA );
  ROMByteWrite( inRomInfo, 0x2AAA, 0x55 );
  ROMByteWrite( inRomInfo, 0x5555, inCommand );
}





/*______________________________________________________________________________

BeginROM

*/

void
BeginROM(
  TRomInfo *		inRomInfo)
{
  FxU32
    theMiscInit1;

  // Remap trick
  ExpMgrConfigReadLong( inRomInfo->mRegEntryID, (LogicalAddress)0x30, &inRomInfo->mOriginalRomBaseAddr);
  inRomInfo->mRomBaseAddr = inRomInfo->mHwBaseAddress + 0xA00000; // Enable ROM accesses
  inRomInfo->mRomBaseAddr |= 1;
  ExpMgrConfigWriteLong( inRomInfo->mRegEntryID, (LogicalAddress)0x30, inRomInfo->mRomBaseAddr);
  inRomInfo->mRomBaseAddr &= ~1;

  HWC_IO_LOAD( inRomInfo->mRegInfo, miscInit1, inRomInfo->mOriginalMiscInit1);
  theMiscInit1 = (inRomInfo->mOriginalMiscInit1 & 0xFDFFFFFF) | 0x10;
  HWC_IO_STORE( inRomInfo->mRegInfo, miscInit1, theMiscInit1);

  inRomInfo->mRomOpen = FXTRUE;
  
  Delay15us(10);            // Wait long enough for the hardware to catch up
}





/*______________________________________________________________________________

EndROM

*/

void
EndROM(
  TRomInfo *		inRomInfo)
{
  ExpMgrConfigWriteLong( inRomInfo->mRegEntryID, (LogicalAddress)0x30, inRomInfo->mOriginalRomBaseAddr);
  HWC_IO_STORE( inRomInfo->mRegInfo, miscInit1, inRomInfo->mOriginalMiscInit1);

  inRomInfo->mRomOpen = FXFALSE;
}





/*______________________________________________________________________________

ROMByteRead

*/

FxU8
ROMByteRead(
  TRomInfo *		inRomInfo,
  FxU16             inOffset)
{
	BYTE	theTemp;

	__dcbf((void *)(inRomInfo->mRomBaseAddr + inOffset),0);
	__sync();
	theTemp = *(volatile BYTE *)(inRomInfo->mRomBaseAddr + inOffset);
	__sync();
	__dcbf((void *)(inRomInfo->mRomBaseAddr + inOffset),0);
	__sync();

	FlushBus( inRomInfo );
	return (theTemp);
}





/*______________________________________________________________________________

ROMByteWrite

*/

void
ROMByteWrite (
  TRomInfo *		inRomInfo,
  FxU16             inOffset,
  FxU8              inData)
{
	__dcbf((void *)(inRomInfo->mRomBaseAddr + inOffset),0);
	__sync();
	*(volatile BYTE *)(inRomInfo->mRomBaseAddr + inOffset) = inData;
	__sync();
	__dcbf((void *)(inRomInfo->mRomBaseAddr + inOffset),0);
	__sync();
	
	FlushBus( inRomInfo );
}





/*______________________________________________________________________________

ROMVerifyWrite

*/

FxU8 ROMVerifyWrite (
  TRomInfo *		inRomInfo,
  FxU16				inOffset,
  FxU8				inExpected)
{
  FxU8				theResult = false;
  FxU8				theData;

  switch ( inRomInfo->mManufacturer )
  {
    case MANID_AMD_TI:
      while (true)
      {
        theData = ROMByteRead( inRomInfo, inOffset );

         // Check that DQ7 is the same as the data
         if (((theData ^ (~inExpected)) & 0x80) == 0x80)
           break;		// Verified!

         // See if DQ5 is a 1 and look for error
         if ((theData & 0x20) == 0x20)
         {
           theData = ROMByteRead( inRomInfo, inOffset );
           // If the data is not there, then there was an error
           if (((theData ^ ~inExpected) & 0x80) == 0x80)
           {
             goto ErrorExit;
           }
         }
       }
       break;


    case MANID_ATMEL:
      switch ( inRomInfo->mDevice )
      {
        case ATMEL_49F010:
          while (true)
          {
            theData = ROMByteRead( inRomInfo, inOffset );

             // Check that DQ7 is the same as the data
             if (((theData ^ (~inExpected)) & 0x80) == 0x80)
               break;		// Verified!
          }
          break;


          default:
            Delay15us (10000/15);
            break;
      }
      break;

    case MANID_SST:
      switch ( inRomInfo->mDevice )
      {
        case SST_29EE010:
        case SST_39SF010:
        case SST_39VF512:
          while (true)
          {
            theData = ROMByteRead( inRomInfo, inOffset );

            // Check that DQ7 is the same as the data
            if (((theData ^ (~inExpected)) & 0x80) == 0x80)
              break;		// Verified!
          }
          break;


        default:

          Delay15us (10000/15);
          break;
      }
      break;
  }

  theResult = true;
  
  
ErrorExit:
	return theResult;
}




/*______________________________________________________________________________

FlushBus

*/

void
FlushBus(
  TRomInfo *		inRomInfo)
{
	// Do two I/O reads
	volatile long foo;
	foo = *((volatile long *)inRomInfo->mRegInfo.ioPortBase);
	foo = *((volatile long *)inRomInfo->mRegInfo.ioPortBase);
}





/*______________________________________________________________________________

Delay15us - inCount x 15 micro second delay

*/

void
Delay15us(
  int		        inCount)
{
  AbsoluteTime delayTime;						// The duration of delay in # ticks
	
  while(inCount > 0)
  {
    delayTime = DurationToAbsolute(-15);
    DelayForHardware(delayTime);
    inCount--;
  }
}


