/******************************Module*Header*******************************\
* Module Name: fastfill.c
*
* Fast routine for drawing polygons that aren't complex in shape.
*
* Copyright (c) 1993-1996 Microsoft Corporation
* Copyright (c) 1997 3Dfx Interactive, Inc.
\**************************************************************************/

#include "precomp.h"

#define RIGHT 0
#define LEFT  1

typedef struct _TRAPEZOIDDATA TRAPEZOIDDATA;    // Handy forward declaration

typedef VOID (FNTRAPEZOID)(TRAPEZOIDDATA*, LONG, LONG);
                                                // Prototype for trapezoid
                                                //   drawing routines

typedef struct _EDGEDATA {
LONG      x;                // Current x position
LONG      dx;               // # pixels to advance x on each scan
LONG      lError;           // Current DDA error
LONG      lErrorUp;         // DDA error increment on each scan
LONG      dN;               // Signed delta-y in fixed point form (also known
                            //   as the DDA error adjustment, and used to be
                            //   called 'lErrorDown')
LONG      dM;               // Signed delta-x in fixed point form
POINTFIX* pptfx;            // Points to start of current edge
LONG      dptfx;            // Delta (in bytes) from pptfx to next point
LONG      cy;               // Number of scans to go for this edge
LONG      bNew;             // Set to TRUE when a new DDA must be started
                            //   for the edge.
} EDGEDATA;                         /* ed, ped */

typedef struct _TRAPEZOIDDATA {
FNTRAPEZOID*    pfnTrap;    // Pointer to appropriate trapezoid drawing routine,
                            //   or trapezoid clip routine
FNTRAPEZOID*    pfnTrapClip;// Pointer to appropriate trapezoid drawing routine
                            //   if doing clipping
PDEV*           ppdev;      // Pointer to PDEV
EDGEDATA        aed[2];     // DDA information for both edges
POINTL          ptlBrush;   // Brush alignment
LONG            yClipTop;   // Top of clip rectangle
LONG            yClipBottom;// Bottom of clip rectangle

} TRAPEZOIDDATA;                    /* td, ptd */

/******************************Public*Routine******************************\
* VOID vClipTrapezoid
*
* Clips a trapezoid.
*
* NOTE: This routine assumes that the polygon's dimensions are small
*       enough that its QUOTIENT_REMAINDER calculations won't overflow.
*       This means that large polygons must never make it here.
*
\**************************************************************************/

VOID vClipTrapezoid(
TRAPEZOIDDATA*  ptd,
LONG            yTrapTop,
LONG            cyTrapezoid)
{
    LONG    yTrapBottom;
    LONG    dN;
    LONG    lNum;
    LONG    xDelta;
    LONG    lError;

    yTrapBottom = yTrapTop + cyTrapezoid;

    if (yTrapTop < ptd->yClipTop)
    {
        if ((ptd->aed[LEFT].bNew) &&
            (yTrapBottom + ptd->aed[LEFT].cy > ptd->yClipTop))
        {
            dN   = ptd->aed[LEFT].dN;
            lNum = ptd->aed[LEFT].dM * (ptd->yClipTop - yTrapTop)
                 + (ptd->aed[LEFT].lError + dN);

            if (lNum >= 0)
            {
                QUOTIENT_REMAINDER(lNum, dN, xDelta, lError);
            }
            else
            {
                lNum = -lNum;

                QUOTIENT_REMAINDER(lNum, dN, xDelta, lError);

                xDelta = -xDelta;
                if (lError != 0)
                {
                    xDelta--;
                    lError = dN - lError;
                }
            }

            ptd->aed[LEFT].x     += xDelta;
            ptd->aed[LEFT].lError = lError - dN;
        }

        if ((ptd->aed[RIGHT].bNew) &&
            (yTrapBottom + ptd->aed[RIGHT].cy > ptd->yClipTop))
        {
            dN   = ptd->aed[RIGHT].dN;
            lNum = ptd->aed[RIGHT].dM * (ptd->yClipTop - yTrapTop)
                 + (ptd->aed[RIGHT].lError + dN);

            if (lNum >= 0)
            {
                QUOTIENT_REMAINDER(lNum, dN, xDelta, lError);
            }
            else
            {
                lNum = -lNum;

                QUOTIENT_REMAINDER(lNum, dN, xDelta, lError);

                xDelta = -xDelta;
                if (lError != 0)
                {
                    xDelta--;
                    lError = dN - lError;
                }
            }

            ptd->aed[RIGHT].x     += xDelta;
            ptd->aed[RIGHT].lError = lError - dN;
        }
    }

    // If this trapezoid vertically intersects our clip rectangle, draw it:

    if ((yTrapBottom > ptd->yClipTop) &&
        (yTrapTop    < ptd->yClipBottom))
    {
        if (yTrapTop <= ptd->yClipTop)
        {
            yTrapTop = ptd->yClipTop;

            // Have to let trapezoid drawer know that it has to load
            // its DDAs for very first trapezoid drawn:

            ptd->aed[RIGHT].bNew = TRUE;
            ptd->aed[LEFT].bNew  = TRUE;
        }

        if (yTrapBottom >= ptd->yClipBottom)
        {
            yTrapBottom = ptd->yClipBottom;
        }

        ptd->pfnTrapClip(ptd, yTrapTop, yTrapBottom - yTrapTop);
    }
}

/******************************Public*Routine******************************\
* VOID vHardwarePolygon
*
* Uses Banshee's hardware polygon capability to draw integer coordinate
* polygons.
*
\**************************************************************************/

VOID vHardwarePolygon(
TRAPEZOIDDATA*  ptd,
LONG            cEdges,      // Number of vertices in vertex array.
POINTFIX*       pptfxFirst,  // Points to the first point in the polygon array
POINTFIX*       pptfxLast,   // Points to the last point in the polygon array
POINTFIX*       pptfxTop)    // Points to the top-most point in the polygon array
{
    PDEV*       ppdev;
    BYTE*       pjH3Base;
    LONG        xOffset;
    LONG        yOffset;
    POINTFIX*   pptfxLeft;      // Points to the last left vertex sent.
    POINTFIX*   pptfxRight;     // Points to the last left vertex sent.
    POINTFIX*   pptfxNextLeft;  // Points to the next left vertex to send.
    POINTFIX*   pptfxNextRight; // Points to the next right vertex to send.
    LONG        lLeftDelta;     // Size delta to go "left" in vertex array.
    LONG        lRightDelta;    // Size delta to go "right" in vertex array.
    LONG        cVertices;      // Number of vertices left to send to HW.
    ULONG       ulIPoint;       // Integer point in Banshee yx format.

    GWH_DECL;
    
    ppdev = ptd->ppdev;
    pjH3Base = ppdev->pjH3Base;

    GWH_PROLOG;

    xOffset     = ppdev->xOffset;
    yOffset     = ppdev->yOffset;

    /*
     *  Banshee HW polygon algorithm.
     *
     *    1. Start at top
     *
     *    2. Move top pointer left one if the y value of the point to the left
     *       of the topmost point equals the y value of the topmost point.
     *       (Check wrap in point array).
     *
     *       This is so that if there is more than one point that shares the
     *       lowest Y coordinate we will always send the vertices to the HW
     *       starting with the leftmost point.  We only have to check one
     *       point to the left because our test for hardware drawing rejected
     *       colinear points.
     *
     *       Note that to step left in the array you add td.aed[LEFT].dptfx.
     *       This is because the polygon can have its vertices ordered in
     *       either a clockwise or counterclockwise orientation and we don't
     *       have to separate the left/right cases this way.
     *
     *    3. Send pptfxTop (the topmost, leftmost point) to srcXY.
     *
     *    4. If more than one point shares the lowest y value (topmost) send
     *       the rightmost top point to dstXY, otherwise send pptfxTop to
     *       dstXY.
     *
     *    5. Initiate command.
     *
     *    6. Keep track of the last y value sent for the left and right sides.
     *       If the y value for the last vertex sent for the left side is
     *       <= the last y value sent for the right side, the next vertex on
     *       the left side should be written to the launch area.  Otherwise,
     *       the next vertex for the right side should be written to the
     *       launch area.
     *
     *    7. When we send the last vertex to the launch area:
     *
     *           // Compare y values of last left and right vertices sent.
     *
     *           if( lvertex.y < rvertex.y )
     *              send rvertex again;
     *           else if( lvertex.y > rvertex.y )
     *              send lvertex again;
     *           else  // lvertex.y == rvertex.y
     *              do nothing -- done;
     *                   
     */


        cVertices = cEdges;

        lLeftDelta  = ptd->aed[LEFT].dptfx;
        lRightDelta = ptd->aed[RIGHT].dptfx;

        pptfxLeft = pptfxTop;
        pptfxRight = pptfxTop;

        // Find the next left vertex, accounting for wrapping:

        pptfxNextLeft = (POINTFIX*) ((BYTE*) pptfxLeft + lLeftDelta);
        if (pptfxNextLeft < pptfxFirst)
            pptfxNextLeft = pptfxLast;
        else if (pptfxNextLeft > pptfxLast)
            pptfxNextLeft = pptfxFirst;

        // Find the next right vertex, accounting for wrapping:

        pptfxNextRight = (POINTFIX*) ((BYTE*) pptfxRight + lRightDelta);
        if (pptfxNextRight < pptfxFirst)
            pptfxNextRight = pptfxLast;
        else if (pptfxNextRight > pptfxLast)
            pptfxNextRight = pptfxFirst;


        // Account for multiple top points.
        
        if( pptfxNextLeft->y == pptfxLeft->y )
        {
            pptfxLeft = pptfxNextLeft;
            cVertices--;
        }
        else if ( pptfxNextRight->y == pptfxRight->y )
        {
            pptfxRight = pptfxNextRight;
            cVertices--;
        }

        cVertices--;        // Account for base top point.

        CHECK_FIFO_ROOM( ppdev, 3 );
        
        GWH_BEGIN_2D_PACKET(3, SSTCP_PKT2_SRCXY |
                               SSTCP_PKT2_DSTXY |
                               SSTCP_PKT2_COMMAND);

        CONVERT_POINT_TO_INTEGER( ulIPoint, pptfxLeft, xOffset, yOffset );
        SET(1, pjH3Base, srcXY, ulIPoint );

        CONVERT_POINT_TO_INTEGER( ulIPoint, pptfxRight, xOffset, yOffset );
        SET(2, pjH3Base, dstXY, ulIPoint );

        SET(3, pjH3Base, command, ppdev->ulCmd      |
                               SSTG_POLYFILL );

        GWH_END_2D_PACKET( 3 );
        
        while( cVertices-- > 0 )
        {
            if( pptfxLeft->y <= pptfxRight->y )
            {
                // Advance left vertex.
                GET_NEXT_VERTEX( pptfxLeft, lLeftDelta );
                CONVERT_POINT_TO_INTEGER( ulIPoint, pptfxLeft,
                                           xOffset, yOffset );
            }
            else
            {
                // Advance right vertex.
                GET_NEXT_VERTEX( pptfxRight, lRightDelta );
                CONVERT_POINT_TO_INTEGER( ulIPoint, pptfxRight,
                                          xOffset, yOffset );
            }

            CHECK_FIFO_ROOM( ppdev, 1 );
            GWH_BEGIN_PKT1_PACKET(1, SSTCP_PKT1_LAUNCH, SSTCP_PKT1_NOINC);
            SET_PKT1_REG(pjH3Base, launch, ulIPoint );
            GWH_END_PKT1_PACKET(1);
        }

        // We are done.  But, check to see if we need to send an additional
        // vertex to flush the poly engine.
        
        if( pptfxLeft->y < pptfxRight->y )
        {
            GET_NEXT_VERTEX( pptfxLeft, lLeftDelta );
            CONVERT_POINT_TO_INTEGER( ulIPoint, pptfxLeft, xOffset, yOffset );
            
            CHECK_FIFO_ROOM( ppdev, 1 );
            GWH_BEGIN_PKT1_PACKET(1, SSTCP_PKT1_LAUNCH, SSTCP_PKT1_NOINC);
            SET_PKT1_REG(pjH3Base, launch, ulIPoint );
            GWH_END_PKT1_PACKET(1);
        }
        else if( pptfxLeft->y > pptfxRight->y )
        {
            GET_NEXT_VERTEX( pptfxRight, lRightDelta );
            CONVERT_POINT_TO_INTEGER( ulIPoint, pptfxRight, xOffset, yOffset );
            
            CHECK_FIFO_ROOM( ppdev, 1 );
            GWH_BEGIN_PKT1_PACKET(1, SSTCP_PKT1_LAUNCH, SSTCP_PKT1_NOINC);
            SET_PKT1_REG(pjH3Base, launch, ulIPoint );
            GWH_END_PKT1_PACKET(1);
        }
        GWH_EPILOG;
}

/******************************Public*Routine******************************\
* VOID vSoftwareTrapezoid
*
* Draws a trapezoid using a software DDA.
*
\**************************************************************************/

VOID vSoftwareTrapezoid(
TRAPEZOIDDATA*  ptd,
LONG            yTrapezoid,
LONG            cyTrapezoid)
{
    PDEV*       ppdev;
    BYTE*       pjH3Base;
    LONG        xOffset;
    LONG        lLeftError;
    LONG        xLeft;
    LONG        lRightError;
    LONG        xRight;
    LONG        lTmp;
    LONG        yDest;
    EDGEDATA    edTmp;

    GWH_DECL;

    ppdev = ptd->ppdev;
    pjH3Base = ppdev->pjH3Base;

    GWH_PROLOG;
    
    xOffset     = ppdev->xOffset;
    yTrapezoid += ppdev->yOffset;

    // If the left and right edges are vertical, simply output as
    // a rectangle:

    if (((ptd->aed[LEFT].lErrorUp | ptd->aed[RIGHT].lErrorUp) == 0) &&
        ((ptd->aed[LEFT].dx       | ptd->aed[RIGHT].dx) == 0))
    {
        /////////////////////////////////////////////////////////////////
        // Vertical-edge special case

        xLeft  = ptd->aed[LEFT].x + xOffset;
        xRight = ptd->aed[RIGHT].x + xOffset;
        if (xLeft > xRight)
        {
            SWAP(xLeft,          xRight,          lTmp);
            SWAP(ptd->aed[LEFT], ptd->aed[RIGHT], edTmp);
        }

        if (xLeft < xRight)
        {

			CHECK_FIFO_ROOM( ppdev, 3);
			
            GWH_BEGIN_2D_PACKET(3, SSTCP_PKT2_DSTSIZE |
                                   SSTCP_PKT2_DSTXY |
                                   SSTCP_PKT2_COMMAND);

	        // Already absolute
	        SET(1, pjH3Base, dstSize, H3_PACKXY_FAST(xRight - xLeft, cyTrapezoid));

    	    SET(2, pjH3Base, dstXY, H3_PACKXY(xLeft, yTrapezoid));

        	SET(3, pjH3Base, command, ppdev->ulCmd      |
                                   SSTG_GO           |
                                   SSTG_RECTFILL );

            GWH_END_2D_PACKET( 3 );

//            MM_MIN_AXIS_PCNT(ppdev, pjMmBase, 0);
        }
    }
    else
    {
//        MM_ABS_DEST_Y(ppdev, pjMmBase, yTrapezoid);
		yDest = yTrapezoid;

        yTrapezoid += cyTrapezoid + 1; // One past end scan

        lLeftError  = ptd->aed[LEFT].lError;
        xLeft       = ptd->aed[LEFT].x + xOffset;
        lRightError = ptd->aed[RIGHT].lError;
        xRight      = ptd->aed[RIGHT].x + xOffset;

        while (TRUE)
        {
            /////////////////////////////////////////////////////////////////
            // Run the DDAs

            if (xLeft < xRight)
            {
                CHECK_FIFO_ROOM( ppdev, 3);
			
                GWH_BEGIN_2D_PACKET(3, SSTCP_PKT2_DSTSIZE |
                                       SSTCP_PKT2_DSTXY |
                                       SSTCP_PKT2_COMMAND);

				// Always set minor_axis_pcnt = 1 pixel.
		        SET(1, pjH3Base, dstSize, H3_PACKXY_FAST(xRight - xLeft, 1) );

		        // Already absolute
    		    SET(2, pjH3Base, dstXY, H3_PACKXY(xLeft, yDest++) );

                SET(3, pjH3Base, command, ppdev->ulCmd      |
                                       SSTG_GO           |
                                       SSTG_RECTFILL );
                GWH_END_2D_PACKET( 3 );

            }
            else if (xLeft > xRight)
            {
                // We don't bother optimizing this case because we should
                // rarely get self-intersecting polygons (if we're slow,
                // the app gets what it deserves).

                SWAP(xLeft,          xRight,          lTmp);
                SWAP(lLeftError,     lRightError,     lTmp);
                SWAP(ptd->aed[LEFT], ptd->aed[RIGHT], edTmp);
                continue;
            }
            else
            {
//                MM_ABS_DEST_Y(ppdev, pjMmBase, yTrapezoid - cyTrapezoid);
				  yDest = yTrapezoid - cyTrapezoid;
            }

            // Advance the right wall:

            xRight      += ptd->aed[RIGHT].dx;
            lRightError += ptd->aed[RIGHT].lErrorUp;

            if (lRightError >= 0)
            {
                lRightError -= ptd->aed[RIGHT].dN;
                xRight++;
            }

            // Advance the left wall:

            xLeft      += ptd->aed[LEFT].dx;
            lLeftError += ptd->aed[LEFT].lErrorUp;

            if (lLeftError >= 0)
            {
                lLeftError -= ptd->aed[LEFT].dN;
                xLeft++;
            }

            cyTrapezoid--;
            if (cyTrapezoid == 0)
                break;
        }

        ptd->aed[LEFT].lError  = lLeftError;
        ptd->aed[LEFT].x       = xLeft - xOffset;
        ptd->aed[RIGHT].lError = lRightError;
        ptd->aed[RIGHT].x      = xRight - xOffset;
    }
    GWH_EPILOG;
}

/******************************Public*Routine******************************\
* VOID vTrapezoidSetup
*
* Initialize the hardware and some state for doing trapezoids.
*
\**************************************************************************/

VOID vTrapezoidSetup(
    PDEV*           ppdev,
    ULONG           rop4,
    ULONG           iSolidColor,
    RBRUSH*         prb,
    POINTL*         pptlBrush,
    TRAPEZOIDDATA*  ptd,
    RECTL*          prclClip,           // NULL if no clipping
    BOOL			bHardwarePolygon) // True if we can use HW poly engine.
{
    ULONG       xyShift;
//    BRUSHENTRY* pbe;
    BYTE*       pjH3Base = ppdev->pjH3Base;
    BYTE		rop3;

    GWH_DECL;
    GWH_PROLOG;

    ptd->ppdev   = ppdev;
    rop3         = (BYTE) rop4;

    ptd->pfnTrap = vSoftwareTrapezoid;  // Assume for now.

    // We filtered out true rop4's in fillpath.c because Banshee doesn't
    // support transparent patterns with arbitrary rops.
    
    if (iSolidColor != -1)
    {
        /////////////////////////////////////////////////////////////////
        // Setup the hardware for solid colours

        if (rop4 == 0xf0f0)
        {
            // Change PATCOPY to SRCCOPY to leverage HW capabilities.
            // Use reversible bit to get NT compatable lines on edges.
            // Add in command type later depending on whether we use
            // SW or HW routine.
            
            ppdev->ulCmd = (0xcc << SSTG_ROP0_SHIFT)    |
                            SSTG_MONO_PATTERN           |
                            SSTG_REVERSIBLE;
                            
	        CHECK_FIFO_ROOM(ppdev, 1);
            GWH_BEGIN_2D_PACKET(1, SSTCP_PKT2_COLORFORE);
                                   
            SET(1, pjH3Base, colorFore, iSolidColor);

            GWH_END_2D_PACKET( 1 );
        }
        else
        {
            ppdev->ulCmd = (rop3 << SSTG_ROP0_SHIFT)    |
                            SSTG_MONO_PATTERN           |
                            SSTG_REVERSIBLE;

	        CHECK_FIFO_ROOM(ppdev, 3);
            GWH_BEGIN_2D_PACKET(3, SSTCP_PKT2_PATTERN0ALIAS |
                                   SSTCP_PKT2_PATTERN1ALIAS |
                                   SSTCP_PKT2_COLORFORE);

            SET(1, pjH3Base, pattern0alias, 0xffffffff );
            SET(2, pjH3Base, pattern1alias, 0xffffffff );
            SET(3, pjH3Base, colorFore, iSolidColor);
            
            GWH_END_2D_PACKET( 3 );
        }
    }
    else
    {
        /////////////////////////////////////////////////////////////////
        // Setup for patterns

//        pbe = prb->apbe[IBOARD(ppdev)];

        xyShift = (  ((-(pptlBrush->y + ppdev->yOffset) & 7) << 3) |
                      (-(pptlBrush->x + ppdev->xOffset) & 7)       )
                         << SSTG_X_PATOFFSET_SHIFT;

        // Pattern fill.
        if (prb->fl & RBRUSH_2COLOR)
        {
            // Monochrome brush.
                ppdev->ulCmd = (rop3 << SSTG_ROP0_SHIFT)    |
                                xyShift                     |
                                SSTG_MONO_PATTERN           |
                                SSTG_REVERSIBLE;
        }
        else
        {
            // Color brush.
            ptd->ptlBrush = *pptlBrush;

            ppdev->ulCmd = (rop3 << SSTG_ROP0_SHIFT)    |
                            xyShift                     |
                            SSTG_REVERSIBLE;

        }
            GWH_EPILOG;
            vH3LoadPatternRegisters( ppdev, prb, FALSE );
            GWH_PROLOG;
    }

    GWH_EPILOG;

    if (prclClip != NULL)
    {
        ptd->pfnTrapClip = ptd->pfnTrap;
        ptd->pfnTrap     = vClipTrapezoid;
        ptd->yClipTop    = prclClip->top;
        ptd->yClipBottom = prclClip->bottom;
        
      	if( prclClip->top < 0 )
   	  		prclClip->top = 0;
      		
   		if( prclClip->bottom > ppdev->cyMemory )
   			prclClip->bottom = ppdev->cyMemory;

        // Uses command fifo.
      	vSetClipping(ppdev, prclClip );
    }
}

/******************************Public*Routine******************************\
* BOOL bFastFill
*
* Draws a non-complex, unclipped polygon.  'Non-complex' is defined as
* having only two edges that are monotonic increasing in 'y'.  That is,
* the polygon cannot have more than one disconnected segment on any given
* scan.  Note that the edges of the polygon can self-intersect, so hourglass
* shapes are permissible.  This restriction permits this routine to run two
* simultaneous DDAs, and no sorting of the edges is required.
*
* Note that NT's fill convention is different from that of Win 3.1 or Win95.
* With the additional complication of fractional end-points, our convention
* is the same as in 'X-Windows'.  But a DDA is a DDA is a DDA, so once you
* figure out how we compute the DDA terms for NT, you're golden.
*
* This routine handles patterns only when the S3 hardware patterns can be
* used.  The reason for this is that once the S3 pattern initialization is
* done, pattern fills appear to the programmer exactly the same as solid
* fills (with the slight difference that different registers and commands
* are used).  Handling 'vIoFillPatSlow' style patterns in this routine
* would be non-trivial...
*
* We take advantage of the fact that the S3 automatically advances the
* current 'y' to the following scan whenever a rectangle is output so that
* we have to write to the accelerator three times for every scan: one for
* the new 'x', one for the new 'width', and one for the drawing command.
*
* Returns TRUE if the polygon was drawn; FALSE if the polygon was complex.
*
\**************************************************************************/

BOOL bFastFill(
PDEV*       ppdev,
LONG        cEdges,         // Includes close figure edge
POINTFIX*   pptfxFirst,
ULONG       rop4,
ULONG       iSolidColor,
RBRUSH*     prb,
POINTL*     pptlBrush,
RECTL*      prclClip)       // NULL if no clipping
{
    LONG      yTrapezoid;   // Top scan for next trapezoid
    LONG      cyTrapezoid;  // Number of scans in current trapezoid
    LONG      yStart;       // y-position of start point in current edge
    LONG      dM;           // Edge delta in FIX units in x direction
    LONG      dN;           // Edge delta in FIX units in y direction
    LONG      i;
    LONG      lCross;       // Cross-product result
    POINTFIX* pptfxLast;    // Points to the last point in the polygon array
    POINTFIX* pptfxTop;     // Points to the top-most point in the polygon
    POINTFIX* pptfxOld;     // Start point in current edge
    POINTFIX* pptfxScan;    // Current edge pointer for finding pptfxTop
    LONG      cScanEdges;   // Number of edges scanned to find pptfxTop
                            //  (doesn't include the closefigure edge)
    LONG      iEdge;
    LONG      lQuotient;
    LONG      lRemainder;
    BYTE*     pjBase;

    TRAPEZOIDDATA   td;     // Edge data and stuff
    EDGEDATA*       ped;    // Points to current edge being processed
    BOOL			bHardwarePolygon;

    /////////////////////////////////////////////////////////////////
    // See if the polygon is convex

    pptfxScan = pptfxFirst;
    pptfxTop  = pptfxFirst;                 // Assume for now that the first
                                            //  point in path is the topmost
    pptfxLast = pptfxFirst + cEdges - 1;

    // Watch for close figure points, because we have the later restriction
    // that we won't allow coincident vertices:

    if ((pptfxLast->x == pptfxFirst->x) && (pptfxLast->y == pptfxFirst->y))
    {
        pptfxLast--;
        cEdges--;
    }

    if (cEdges <= 2)
        goto ReturnTrue;

    // 'pptfxScan' will always point to the first point in the current
    // edge, and 'cScanEdges' will the number of edges remaining, including
    // the current one:

    cScanEdges = cEdges - 1;     // The number of edges, not counting close figure

    if ((pptfxScan + 1)->y > pptfxScan->y)
    {
        // Collect all downs:

        do {
            if (--cScanEdges == 0)
                goto SetUpForFilling;
            pptfxScan++;
        } while ((pptfxScan + 1)->y >= pptfxScan->y);

        // Collect all ups:

        do {
            if (--cScanEdges == 0)
                goto SetUpForFillingCheck;
            pptfxScan++;
        } while ((pptfxScan + 1)->y <= pptfxScan->y);

        // Collect all downs:

        pptfxTop = pptfxScan;

        do {
            if ((pptfxScan + 1)->y > pptfxFirst->y)
                break;

            if (--cScanEdges == 0)
                goto SetUpForFilling;
            pptfxScan++;
        } while ((pptfxScan + 1)->y >= pptfxScan->y);

        goto ReturnFalse;
    }
    else
    {
        // Collect all ups:

        do {
            pptfxTop++;                 // We increment this now because we
                                        //  want it to point to the very last
                                        //  point if we early out in the next
                                        //  statement...
            if (--cScanEdges == 0)
                goto SetUpForFilling;
        } while ((pptfxTop + 1)->y <= pptfxTop->y);

        // Collect all downs:

        pptfxScan = pptfxTop;
        do {
            if (--cScanEdges == 0)
                goto SetUpForFilling;
            pptfxScan++;
        } while ((pptfxScan + 1)->y >= pptfxScan->y);

        // Collect all ups:

        do {
            if ((pptfxScan + 1)->y < pptfxFirst->y)
                break;

            if (--cScanEdges == 0)
                goto SetUpForFilling;
            pptfxScan++;
        } while ((pptfxScan + 1)->y <= pptfxScan->y);

        goto ReturnFalse;
    }

SetUpForFillingCheck:

    // We check to see if the end of the current edge is higher
    // than the top edge we've found so far:

    if ((pptfxScan + 1)->y < pptfxTop->y)
        pptfxTop = pptfxScan + 1;

SetUpForFilling:

    bHardwarePolygon = TRUE;	// Assume for now we can use HW trapezoid.
  
  	// First check to see if the trapezoid consists of all integer
  	// coordinates.  Fall back to Software Trapezoid if not integers.
  	
	for( pptfxScan = pptfxFirst; pptfxScan <= pptfxLast; pptfxScan++ )
	{
		if(  (pptfxScan->x | pptfxScan->y) & 0xf  )
		{
			bHardwarePolygon = FALSE;
			goto FinishedHWTrapCheck;
		}
	}
  
    pptfxScan  = pptfxFirst;
    cScanEdges = cEdges - 2;

    // NOTE: For a bit of speed and simplicity, we will assume that
    //       our cross product calculations will not overflow.  A
    //       consequence of this is that the caller MUST ensure that
    //       the bounds of the polygon are small enough that there
    //       will be no overflow.

    lCross = (((pptfxScan + 1)->x - (pptfxScan + 0)->x)
            * ((pptfxScan + 2)->y - (pptfxScan + 1)->y)
            - ((pptfxScan + 1)->y - (pptfxScan + 0)->y)
            * ((pptfxScan + 2)->x - (pptfxScan + 1)->x));

    if (lCross == 0)
    {
        // We don't allow any colinear points into FastFill.  We do this
        // here because we would need a non-zero cross product to determine
        // which direction the rest of the edges should go.  We do this
        // later so that coincident vertices will never mess us up by
        // hiding a cross product sign change.

		bHardwarePolygon = FALSE;
        goto FinishedHWTrapCheck;
    }
    else if (lCross > 0)
    {
        // Make sure all cross products are positive:

        pptfxScan++;
        while (--cScanEdges != 0)
        {
            if (((pptfxScan + 1)->x - (pptfxScan + 0)->x)
              * ((pptfxScan + 2)->y - (pptfxScan + 1)->y)
              - ((pptfxScan + 1)->y - (pptfxScan + 0)->y)
              * ((pptfxScan + 2)->x - (pptfxScan + 1)->x) <= 0)
            {
				bHardwarePolygon = FALSE;
                goto FinishedHWTrapCheck;
            }
            pptfxScan++;
        }

        // Check the angles formed by the closefigure edge:

        if (((pptfxScan + 1)->x - (pptfxScan + 0)->x)
          * ((pptfxFirst   )->y - (pptfxScan + 1)->y)
          - ((pptfxScan + 1)->y - (pptfxScan + 0)->y)
          * ((pptfxFirst   )->x - (pptfxScan + 1)->x) <= 0)
        {
			bHardwarePolygon = FALSE;
            goto FinishedHWTrapCheck;
        }

        if (((pptfxFirst    )->x - (pptfxScan + 1)->x)
          * ((pptfxFirst + 1)->y - (pptfxFirst   )->y)
          - ((pptfxFirst    )->y - (pptfxScan + 1)->y)
          * ((pptfxFirst + 1)->x - (pptfxFirst   )->x) <= 0)
        {
			bHardwarePolygon = FALSE;
            goto FinishedHWTrapCheck;
        }

        // The figure has its points ordered in a clockwise direction:

        td.aed[LEFT].dptfx  = -(LONG) sizeof(POINTFIX);
        td.aed[RIGHT].dptfx = sizeof(POINTFIX);
    }
    else
    {
        // Make sure all cross products are negative:

        pptfxScan++;
        while (--cScanEdges != 0)
        {
            if (((pptfxScan + 1)->x - (pptfxScan + 0)->x)
              * ((pptfxScan + 2)->y - (pptfxScan + 1)->y)
              - ((pptfxScan + 1)->y - (pptfxScan + 0)->y)
              * ((pptfxScan + 2)->x - (pptfxScan + 1)->x) >= 0)
            {
				bHardwarePolygon = FALSE;
                goto FinishedHWTrapCheck;
            }
            pptfxScan++;
        }

        // Check the angles formed by the closefigure edge:

        if (((pptfxScan + 1)->x - (pptfxScan + 0)->x)
          * ((pptfxFirst   )->y - (pptfxScan + 1)->y)
          - ((pptfxScan + 1)->y - (pptfxScan + 0)->y)
          * ((pptfxFirst   )->x - (pptfxScan + 1)->x) >= 0)
        {
			bHardwarePolygon = FALSE;
            goto FinishedHWTrapCheck;
        }

        if (((pptfxFirst    )->x - (pptfxScan + 1)->x)
          * ((pptfxFirst + 1)->y - (pptfxFirst   )->y)
          - ((pptfxFirst    )->y - (pptfxScan + 1)->y)
          * ((pptfxFirst + 1)->x - (pptfxFirst   )->x) >= 0)
        {
			bHardwarePolygon = FALSE;
            goto FinishedHWTrapCheck;
        }

FinishedHWTrapCheck:
        // The figure has its points ordered in a counter-clockwise direction:
        // (Guess as to the ordering of the points for vSoftwareTrapezoid).

        td.aed[LEFT].dptfx  = sizeof(POINTFIX);
        td.aed[RIGHT].dptfx = -(LONG) sizeof(POINTFIX);
    }

    /////////////////////////////////////////////////////////////////
    // Some Initialization

    td.aed[LEFT].pptfx  = pptfxTop;
    td.aed[RIGHT].pptfx = pptfxTop;

    yTrapezoid = (pptfxTop->y + 15) >> 4;

    // Make sure we initialize the DDAs appropriately:

    td.aed[LEFT].cy  = 0;
    td.aed[RIGHT].cy = 0;

    vTrapezoidSetup(ppdev, rop4, iSolidColor, prb, pptlBrush, &td,
                    prclClip, bHardwarePolygon);

    if( bHardwarePolygon )
    {
        vHardwarePolygon( &td, cEdges, pptfxFirst, pptfxLast, pptfxTop );
        goto ResetClippingAndReturnTrue;
    }
    
NewTrapezoid:

    /////////////////////////////////////////////////////////////////
    // DDA initialization

    for (iEdge = 1; iEdge >= 0; iEdge--)
    {
        ped       = &td.aed[iEdge];
        ped->bNew = FALSE;
        if (ped->cy == 0)
        {
            // Our trapezoid drawing routine may want to be notified when
            // it will have to reset its DDA to start a new edge:

            ped->bNew = TRUE;

            // Need a new DDA:

            do {
                cEdges--;
                if (cEdges < 0)
                    goto ResetClippingAndReturnTrue;

                // Find the next left edge, accounting for wrapping:

                pptfxOld = ped->pptfx;
                ped->pptfx = (POINTFIX*) ((BYTE*) ped->pptfx + ped->dptfx);

                if (ped->pptfx < pptfxFirst)
                    ped->pptfx = pptfxLast;
                else if (ped->pptfx > pptfxLast)
                    ped->pptfx = pptfxFirst;

                // Have to find the edge that spans yTrapezoid:

                ped->cy = ((ped->pptfx->y + 15) >> 4) - yTrapezoid;

                // With fractional coordinate end points, we may get edges
                // that don't cross any scans, in which case we try the
                // next one:

            } while (ped->cy <= 0);

            // 'pptfx' now points to the end point of the edge spanning
            // the scan 'yTrapezoid'.

            dN = ped->pptfx->y - pptfxOld->y;
            dM = ped->pptfx->x - pptfxOld->x;

            ASSERTDD(dN > 0, "Should be going down only");

            // Compute the DDA increment terms:

            ped->dM = dM;                   // Not used for software trapezoid

            if (dM < 0)
            {
                dM = -dM;
                if (dM < dN)                // Can't be '<='
                {
                    ped->dx       = -1;
                    ped->lErrorUp = dN - dM;
                }
                else
                {
                    QUOTIENT_REMAINDER(dM, dN, lQuotient, lRemainder);

                    ped->dx       = -lQuotient;     // - dM / dN
                    ped->lErrorUp = lRemainder;     // dM % dN
                    if (ped->lErrorUp > 0)
                    {
                        ped->dx--;
                        ped->lErrorUp = dN - ped->lErrorUp;
                    }
                }
            }
            else
            {
                if (dM < dN)                // Can't be '<='
                {
                    ped->dx       = 0;
                    ped->lErrorUp = dM;
                }
                else
                {
                    QUOTIENT_REMAINDER(dM, dN, lQuotient, lRemainder);

                    ped->dx       = lQuotient;      // dM / dN
                    ped->lErrorUp = lRemainder;     // dM % dN
                }
            }

            ped->dN = dN; // DDA limit
            ped->lError     = -1; // Error is initially zero (add dN - 1 for
                                  //  the ceiling, but subtract off dN so that
                                  //  we can check the sign instead of comparing
                                  //  to dN)

            ped->x = pptfxOld->x;
            yStart = pptfxOld->y;

            if ((yStart & 15) != 0)
            {
                // Advance to the next integer y coordinate

                for (i = 16 - (yStart & 15); i != 0; i--)
                {
                    ped->x      += ped->dx;
                    ped->lError += ped->lErrorUp;
                    if (ped->lError >= 0)
                    {
                        ped->lError -= ped->dN;
                        ped->x++;
                    }
                }
            }

            if ((ped->x & 15) != 0)
            {
                ped->lError -= ped->dN * (16 - (ped->x & 15));
                ped->x += 15;       // We'll want the ceiling in just a bit...
            }

            // Chop off those fractional bits:

            ped->x      >>= 4;
            ped->lError >>= 4;
        }
    }

    cyTrapezoid = min(td.aed[LEFT].cy, td.aed[RIGHT].cy); // # of scans in this trap
    td.aed[LEFT].cy  -= cyTrapezoid;
    td.aed[RIGHT].cy -= cyTrapezoid;

    td.pfnTrap(&td, yTrapezoid, cyTrapezoid);

    yTrapezoid += cyTrapezoid;

    goto NewTrapezoid;

ResetClippingAndReturnTrue:

    if (prclClip != NULL)
    {
    	// Only really need to reset X clipping if we want to save a couple
    	// of instructions.
        vResetClipping(ppdev);
    }

ReturnTrue:

    return(TRUE);

ReturnFalse:

    return(FALSE);
}
