/**************************************************************************
*								          *
* 		 Copyright (C) 1997, Silicon Graphics, Inc.	          *
*								          *
*  These coded instructions, statements, and computer programs  contain   * 
*  unpublished  proprietary  information of Silicon Graphics, Inc., and   *
*  are protected by Federal copyright law.  They  may  not be disclosed   *
*  to  third  parties  or copied or duplicated in any form, in whole or   *
*  in part, without the prior written consent of Silicon Graphics, Inc.   *
*								          *
**************************************************************************/
      
/*
 * This program tests OGL 1.1 polygon offset.
 */
      
#include <assert.h>
#include "ogtst.h"	/* include test environment		*/

#ifdef GL_VERSION_1_1
      
static GLboolean CImode;

#define MIN(a, b) (a) < (b) ? (a) : (b)
#define MAX(a, b) (a) > (b) ? (a) : (b)

/*
** Maximum possible unsigned value for an <a> bit number
*/
#define MAX_BIT_VAL(a)  (1 << (a)) - 1

/*
** Structure to make passing coordinates more convenient
*/
typedef struct {
    GLint x1, y1;
    GLint x2, y2;
    GLint z1, z2, z3, z4;
    GLint minx, maxx; /* for testing reporting, and sorting */
    GLint miny, maxy;
    GLboolean too_close; /* one of the z values is z == 0 */
} Coords;


/*
** Test polygon offsets by drawing two rectangles, both with the
** same x, y, and z coordinates, one with polygon offset enabled,
** one with it disabled. The pgonoffset parameters are chosen to
** either displace the polygon closer or farther from the viewer.
**
** In either case, the code always draws the polygon that should be
** farthest last. this is done so that if offset makes no difference,
** or moves in the wrong direction, the test will catch it, since in
** case of GL_LESS, the last object drawn wins.
**
** NOTE: zero-area polygons are drawn, and tested for background color,
** in order to more fully exercise the code.
*/
static void test_offset_drawing(GLfloat minfactor, GLfloat maxfactor, 
			 GLfloat minUnits, GLfloat maxUnits,
			 Coords *coords,
			 char *logmessage)
{
    int obj;
    GLfloat factor, units;
    GLboolean negative_offset; /* which way are we offsetting? */
    GLuint visible_color, obscured_color;
    ogLibClear(0) ;
    glClearDepth(1.);
    glClear(GL_DEPTH_BUFFER_BIT);
    
    ogEnvLog(1, "%s", logmessage);

    /* 
    ** true if any of the parameters are negative. Assuming that parameters
    ** are all on one side or the other of zero.
    */
    negative_offset = minfactor < 0.f ||
	              maxfactor < 0.f ||
		      minUnits   < 0.f ||
		      maxUnits   < 0.f;


    factor = ogLibFloatRand(minfactor, maxfactor);
    units   = ogLibFloatRand(minUnits, maxUnits);
    ogEnvLog(1, "factor = %.4f, units = %f\n", factor, units);

    START_DL_OR_IM(1);

    glPolygonOffset(factor, units);

    if(negative_offset) /* draw closer pgon first */
	glEnable(GL_POLYGON_OFFSET_FILL);

    visible_color = ogLibColor();

    glBegin(GL_POLYGON);	
    glVertex3i(coords->x1,coords->y1,coords->z1);
    glVertex3i(coords->x1,coords->y2,coords->z2);
    glVertex3i(coords->x2,coords->y2,coords->z3);
    glVertex3i(coords->x2,coords->y1,coords->z4);
    glEnd();


    if(negative_offset)
	glDisable(GL_POLYGON_OFFSET_FILL);
    else
	glEnable(GL_POLYGON_OFFSET_FILL);

    /* if coords->z1 is front of z range, offseting won't obscure polygon.
    ** if this happens, go ahead and run the test, to more fully
    ** exercise the code, but don't check for different colors.
    ** This is done by not changing the color, resulting in both
    ** polygons having the same color.
    */

    /* XXX this shouldn't be exactly zero, but a small number */
    if(coords->too_close) 
	obscured_color = ogLibColor(); /* obscured color */

    ogEnvLog(2, "Obscured Rectangle\n");
    glBegin(GL_POLYGON);	
    glVertex3i(coords->x1,coords->y1,coords->z1);
    glVertex3i(coords->x1,coords->y2,coords->z2);
    glVertex3i(coords->x2,coords->y2,coords->z3);
    glVertex3i(coords->x2,coords->y1,coords->z4);
    glEnd();

    if(!negative_offset)
	glDisable(GL_POLYGON_OFFSET_FILL);

    FINIS_DL_OR_IM(1);

    ogEnvLog(2, "Visible Color = %x; Obscured Color = %x\n",
	     visible_color, obscured_color);

    ogEnvLog(3,"box coords: (%d,%d), (%d,%d)\n", 
	     coords->minx,
	     coords->miny,
	     coords->maxx,
	     coords->maxy);

    /* check the edges */
    if(!(coords->maxx - coords->minx) || 
       !(coords->maxy - coords->miny)) /* zero area pgons draw nothing */
	visible_color = 0; 

    ogLibRectCheck(coords->minx,
		   coords->miny,
		   coords->maxx - 1,
		   coords->maxy - 1,
		   visible_color, 0);
}


/*
** Parameter for controlling the type of z values generated by make_zs.
*/
typedef enum { Nodz,      /* dz is 0 */
	       Dzdx,      /* dz != 0 in dx direction */
	       Dzdy,      /* dz != 0 in dy direction */
	       Dzdxdy,    /* dz != 0 in both dx and dy directions */
	       Nonplanar, /* rectangle isn't planer in z */
	       Random,    /* generate some random z's */
	       Last       /* iteration limit */
	     } Ztype;




/*
** Generate a set of coordinates to generate rectangles with various
** z slopes.
*/
static Coords *make_coords(GLint xmax, GLint ymax, Ztype type)
{
    static Coords coords;

    /* generate x and y values */

    coords.x1 = 2 + ogLibIntRand(0, xmax - 4);
    coords.y1 = 2 + ogLibIntRand(0, ymax - 4);
    coords.x2 = 2 + ogLibIntRand(0, xmax - 4);
    coords.y2 = 2 + ogLibIntRand(0, ymax - 4);

    /* generate z values per type request */

    coords.z1 = 
    coords.z2 =
    coords.z3 =
    coords.z4 = 
    ogLibIntRand(0, MAX_BIT_VAL(24));

    switch(type) {
    case Nodz: /* already have dz == 0 */
	ogEnvLog(1, "***Rectangle in z == 0 plane\n");
	break;
    case Dzdx: /* z1 == z2, z3 == z4 */
	ogEnvLog(1, "***Rectangle with dx != 0\n");
	while(coords.z1 == coords.z3)
	    coords.z3 = ogLibIntRand(0, MAX_BIT_VAL(24));
	coords.z4 = coords.z3;
	break;
    case Dzdy: /* z1 == z4, z2 == z3 */
	ogEnvLog(1, "***Rectangle with dy != 0\n");
	while(coords.z1 == coords.z2)
	    coords.z2 = ogLibIntRand(0, MAX_BIT_VAL(24));
	coords.z3 = coords.z2;
	break;
    case Dzdxdy: /* dz/dx and dz/dy != 0, in planar polygon */
	ogEnvLog(1, "***Planar rectangle with dy and dx != 0 \n");
	while(coords.z1 == coords.z3)
	    coords.z3 = ogLibIntRand(0, MAX_BIT_VAL(24));
	coords.z4 = coords.z1 + ogLibIntRand(0, coords.z3 - coords.z1);
	coords.z2 = coords.z3 - (coords.z4 - coords.z1);
	/* are z2 and z4 reasonable values? */
	assert((coords.z2 <= MAX_BIT_VAL(24)) &&
	       (coords.z2 >= 0)); 
	assert((coords.z4 <= MAX_BIT_VAL(24)) &&
	       (coords.z4 >= 0));
	/*
	** Are all z's in the same plane?
	** if so, z2 and z4 (minus z1) should add to z3 - z1
	** if so, z3 = z1 + (z2 - z1) + (z4 - z1) -> z2 + z4 - z1
	*/
	assert(coords.z3 ==  coords.z2 + coords.z4 - coords.z1);
	break;
    case Nonplanar: /* one vertex sticking out of z == 0 plane */
	ogEnvLog(1, "***Non-planar rectangle with dy and dx != 0 \n");
	while(coords.z1 == coords.z3)
	    coords.z3 = ogLibIntRand(0, MAX_BIT_VAL(24));
	coords.z4 = coords.z3;
	coords.z2 = coords.z3;
	break;
    case Random:
	ogEnvLog(1, "***Non-planar rectangle with random z values\n");
	coords.z2 = ogLibIntRand(0, MAX_BIT_VAL(24));
	coords.z3 = ogLibIntRand(0, MAX_BIT_VAL(24));
	coords.z4 = ogLibIntRand(0, MAX_BIT_VAL(24));
	break;
    default:
	assert(0); /* is <type> one of the legal enumerated values? */
	break;
    }

    coords.minx = MIN(coords.x1, coords.x2);
    coords.maxx = MAX(coords.x1, coords.x2);
    coords.miny = MIN(coords.y1, coords.y2);
    coords.maxy = MAX(coords.y1, coords.y2);

    if(coords.z1 == 0 ||
       coords.z2 == 0 ||
       coords.z3 == 0 ||
       coords.z4 == 0)
	coords.too_close = GL_TRUE;


    /* describe z values generated */

     ogEnvLog(2,  /* rectangle coordinates */
	      "1: %d, %d, %d\n"
	      "2: %d, %d, %d\n"
	      "3: %d, %d, %d\n"
	      "4: %d, %d, %d\n",
	      coords.x1, coords.y1, coords.z1,
	      coords.x1, coords.y2, coords.z2,
	      coords.x2, coords.y2, coords.z3,
	      coords.x2, coords.y1, coords.z4);

    ogEnvLog(2, "dx = %d\n", coords.x2 - coords.x1);
    ogEnvLog(2, "dy = %d\n", coords.y2 - coords.y1);

    ogEnvLog(2, "dz/dx = %.3f\n", 
	     (coords.x2 - coords.x1)/(GLfloat)(coords.z4 - coords.z1));
    ogEnvLog(2, "dz/dy = %.3f\n", 
	     (coords.y2 - coords.y1)/(GLfloat)(coords.z2 - coords.z1));

    return &coords;
}
#endif

TESTMOD(pgonoffset)
{
    Coords *coords;
    GLint xmax, ymax;
    GLint rbits, gbits, bbits, abits;
    Ztype type = Nodz; /* loop through rectangle z types */

#ifdef GL_VERSION_1_1

    if (IS_ONEONE()) {

       xmax = ogEnvQuery(OG_XWSIZE) - 1;
       ymax = ogEnvQuery(OG_YWSIZE) - 1;
       glMatrixMode(GL_PROJECTION);
       glOrtho(0.,(double) xmax+1., 0., (double) ymax+1., 0.,
	       -(double)(MAX_BIT_VAL(24)));

       glGetBooleanv(GL_INDEX_MODE,&CImode); 
       glEnable(GL_DEPTH_TEST);

       /* 
	* try to catch double hits! 
	*/

       if (CImode) 
	   glLogicOp(GL_XOR);
       else
	   glBlendFunc(GL_ONE,GL_ONE);


       glGetIntegerv(GL_RED_BITS, &rbits);
       glGetIntegerv(GL_GREEN_BITS, &gbits);
       glGetIntegerv(GL_BLUE_BITS, &bbits);
       glGetIntegerv(GL_ALPHA_BITS, &abits);

       ogEnvLog(1, "Visual color depths: red %d, green %d, blue %d, alpha %d\n",
		rbits, gbits, bbits, abits);

   /* TODO: check stipple, stencil, stippled stencil */

       while (pass--) {

	   /* Test negative polygon offsets */
	   /* draw a rectangle with 2 poly triangles */
	   ogLibClear(0) ;
	   glClearDepth(1.);
	   glClear(GL_DEPTH_BUFFER_BIT);


	   coords = make_coords(xmax, ymax, type);
	   type = (Ztype) ((type + 1) % (int)Random); /* cycle through most z types */

	   /* test negative units */

	   test_offset_drawing(0.f, 0.f, -0.0, -2.0, coords,
			       "--- negative units ---\n");

	   /* test positive units */

	   test_offset_drawing(0.f, 0.f, 0.0, 100000.0, coords,
			       "--- positive units ---\n");


	   /* test negative factor */

	   test_offset_drawing(-.00001f, -100000.f, 0.f, 0.f, coords,
			       "--- negative factor ---\n");

	   /* test positive factor */

	   test_offset_drawing(0.f, 100000.f, 0.f, 0.f, coords,
			       "--- positive factor ---\n");

       }

    }
#endif
}
      
CLEANUP(pgonoffset)
{
#ifdef GL_VERSION_1_1 

    if (IS_ONEONE()) {
       ogLibSetDefaultColors();
       if (CImode) 
	   glLogicOp(GL_COPY);
       else
	   glBlendFunc(GL_ONE,GL_ZERO);
       glLoadIdentity();
       glMatrixMode(GL_MODELVIEW);
       glDisable(GL_DEPTH_TEST);

       glDisable(GL_POLYGON_OFFSET_FILL);
       glPolygonOffset(0., 0.);

    }

#endif
}
